@@ -1,14 +1,16 @@
 #!/usr/bin/perl -w
 
 # mailgraph -- an rrdtool frontend for mail statistics
-# copyright (c) 2000-2005 David Schweikert <dws@ee.ethz.ch>
+# copyright (c) 2000-2007 ETH Zurich
+# copyright (c) 2000-2007 David Schweikert <david@schweikert.ch>
 # released under the GNU General Public License
 
-######## Parse::Syslog 1.04 (automatically embedded) ########
+######## Parse::Syslog 1.09 (automatically embedded) ########
 package Parse::Syslog;
 use Carp;
 use Symbol;
 use Time::Local;
+use IO::File;
 use strict;
 use vars qw($VERSION);
 my %months_map = (
@@ -21,24 +23,54 @@ my %months_map = (
     'jul' => 6, 'aug' => 7, 'sep' => 8,
     'oct' => 9, 'nov' =>10, 'dec' =>11,
 );
-# year-increment algorithm: if in january, if december is seen, decrement year
-my $enable_year_decrement = 1;
+sub is_dst_switch($$$)
+{
+    my ($self, $t, $time) = @_;
+    # calculate the time in one hour and see if the difference is 3600 seconds.
+    # if not, we are in a dst-switch hour
+    # note that right now we only support 1-hour dst offsets
+    # cache the result
+    if(defined $self->{is_dst_switch_last_hour} and
+        $self->{is_dst_switch_last_hour} == $t->[3]<<5+$t->[2]) {
+        return @{$self->{is_dst_switch_result}};
+    }
+    # calculate a number out of the day and hour to identify the hour
+    $self->{is_dst_switch_last_hour} = $t->[3]<<5+$t->[2];
+    # calculating hour+1 (below) is a problem if the hour is 23. as far as I
+    # know, nobody does the DST switch at this time, so just assume it isn't
+    # DST switch if the hour is 23.
+    if($t->[2]==23) {
+        @{$self->{is_dst_switch_result}} = (0, undef);
+        return @{$self->{is_dst_switch_result}};
+    }
+    # let's see the timestamp in one hour
+    # 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
+    my $time_plus_1h = timelocal($t->[0], $t->[1], $t->[2]+1, $t->[3], $t->[4], $t->[5]);
+    if($time_plus_1h - $time > 4000) {
+        @{$self->{is_dst_switch_result}} = (3600, $time-$time%3600+3600);
+    }
+    else {
+        @{$self->{is_dst_switch_result}} = (0, undef);
+    }
+    return @{$self->{is_dst_switch_result}};
+}
 # fast timelocal, cache minute's timestamp
 # don't cache more than minute because of daylight saving time switch
-my @str2time_last_minute;
-my $str2time_last_minute_timestamp;
 # 0: sec, 1: min, 2: h, 3: day, 4: month, 5: year
-sub str2time($$$$$$$)
+sub str2time($$$$$$$$)
 {
+    my $self = shift @_;
     my $GMT = pop @_;
-    if(defined $str2time_last_minute[4] and
-        $str2time_last_minute[0] == $_[1] and
-        $str2time_last_minute[1] == $_[2] and
-        $str2time_last_minute[2] == $_[3] and
-        $str2time_last_minute[3] == $_[4] and
-        $str2time_last_minute[4] == $_[5])
+    my $lastmin = $self->{str2time_lastmin};
+    if(defined $lastmin and
+        $lastmin->[0] == $_[1] and
+        $lastmin->[1] == $_[2] and
+        $lastmin->[2] == $_[3] and
+        $lastmin->[3] == $_[4] and
+        $lastmin->[4] == $_[5])
     {
-        return $str2time_last_minute_timestamp + $_[0];
+        $self->{last_time} = $self->{str2time_lastmin_time} + $_[0];
+        return $self->{last_time} + ($self->{dst_comp}||0);
     }
     my $time;
     if($GMT) {
@@ -47,9 +79,35 @@ sub str2time($$$$$$$)
     else {
         $time = timelocal(@_);
     }
-    @str2time_last_minute = @_[1..5];
-    $str2time_last_minute_timestamp = $time-$_[0];
-    return $time;
+    # compensate for DST-switch
+    # - if a timewarp is detected (1:00 -> 1:30 -> 1:00):
+    # - test if we are in a DST-switch-hour
+    # - compensate if yes
+    # note that we assume that the DST-switch goes like this:
+    # time   1:00  1:30  2:00  2:30  2:00  2:30  3:00  3:30
+    # stamp   1     2     3     4     3     3     7     8  
+    # comp.   0     0     0     0     2     2     0     0
+    # result  1     2     3     4     5     6     7     8
+    # old Time::Local versions behave differently (1 2  5 6 5 6 7 8)
+    if(!$GMT and !defined $self->{dst_comp} and
+        defined $self->{last_time} and
+        $self->{last_time}-$time > 1200 and
+        $self->{last_time}-$time < 3600)
+    {
+        my ($off, $until) = $self->is_dst_switch(\@_, $time);
+        if($off) {
+            $self->{dst_comp} = $off;
+            $self->{dst_comp_until} = $until;
+        }
+    }
+    if(defined $self->{dst_comp_until} and $time > $self->{dst_comp_until}) {
+        delete $self->{dst_comp};
+        delete $self->{dst_comp_until};
+    }
+    $self->{str2time_lastmin} = [ @_[1..5] ];
+    $self->{str2time_lastmin_time} = $time-$_[0];
+    $self->{last_time} = $time;
+    return $time+($self->{dst_comp}||0);
 }
 sub _use_locale($)
 {
@@ -73,13 +131,25 @@ sub new($$;%)
     }
     $data{type} = 'syslog' unless defined $data{type};
     $data{_repeat}=0;
-    if(ref $file eq 'File::Tail') {
-        $data{filetail} = 1;
+    if(UNIVERSAL::isa($file, 'IO::Handle')) {
         $data{file} = $file;
     }
+    elsif(UNIVERSAL::isa($file, 'File::Tail')) {
+        $data{file} = $file;
+        $data{filetail}=1;
+    }
+    elsif(! ref $file) {
+        if($file eq '-') {
+            my $io = new IO::Handle;
+            $data{file} = $io->fdopen(fileno(STDIN),"r");
+        }
+        else {
+            $data{file} = new IO::File($file, "<");
+            defined $data{file} or croak "can't open $file: $!";
+        }
+    }
     else {
-        $data{file}=gensym;
-        open($data{file}, "<$file") or croak "can't open $file: $!";
+        croak "argument must be either a file-name or an IO::Handle object.";
     }
     if(defined $data{locale}) {
         if(ref $data{locale} eq 'ARRAY') {
@@ -100,15 +170,15 @@ sub _year_increment($$)
     # year change
     if($mon==0) {
         $self->{year}++ if defined $self->{_last_mon} and $self->{_last_mon} == 11;
-        $enable_year_decrement = 1;
+        $self->{enable_year_decrement} = 1;
     }
     elsif($mon == 11) {
-        if($enable_year_decrement) {
+        if($self->{enable_year_decrement}) {
             $self->{year}-- if defined $self->{_last_mon} and $self->{_last_mon} != 11;
         }
     }
     else {
-        $enable_year_decrement = 0;
+        $self->{enable_year_decrement} = 0;
     }
     $self->{_last_mon} = $mon;
 }
@@ -120,7 +190,7 @@ sub _next_line($)
         return $f->read;
     }
     else {
-        return <$f>;
+        return $f->getline;
     }
 }
 sub _next_syslog($)
@@ -130,17 +200,19 @@ sub _next_syslog($)
         $self->{_repeat}--;
         return $self->{_repeat_data};
     }
-    line: while(my $str = $self->_next_line()) {
+    my $file = $self->{file};
+    line: while(defined (my $str = $self->_next_line)) {
         # date, time and host 
         $str =~ /^
-            (\S{3})\s+(\d+)   # date  -- 1, 2
+            (\S{3})\s+(\d+)      # date  -- 1, 2
             \s
-            (\d+):(\d+):(\d+) # time  -- 3, 4, 5
-            (?:\s<\w+\.\w+>)? # FreeBSD's verbose-mode
+            (\d+):(\d+):(\d+)    # time  -- 3, 4, 5
+            (?:\s<\w+\.\w+>)?    # FreeBSD's verbose-mode
             \s
-            ([-\w\.\@:]+)     # host  -- 6
+            ([-\w\.\@:]+)        # host  -- 6
             \s+
-            (.*)              # text  -- 7
+            (?:\[LOG_[A-Z]+\]\s+)?  # FreeBSD
+            (.*)                 # text  -- 7
             $/x or do
         {
             warn "WARNING: line not in syslog format: $str";
@@ -150,7 +222,7 @@ sub _next_syslog($)
         defined $mon or croak "unknown month $1\n";
         $self->_year_increment($mon);
         # convert to unix time
-        my $time = str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
+        my $time = $self->str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
         if(not $self->{allow_future}) {
             # accept maximum one day in the present future
             if($time - time > 86400) {
@@ -218,7 +290,8 @@ sub _next_syslog($)
 sub _next_metalog($)
 {
     my ($self) = @_;
-    line: while(my $str = $self->_next_line()) {
+    my $file = $self->{file};
+    line: while(my $str = $self->_next_line) {
 	# date, time and host 
 	$str =~ /^
             (\S{3})\s+(\d+)   # date  -- 1, 2
@@ -236,7 +309,7 @@ sub _next_metalog($)
         defined $mon or croak "unknown month $1\n";
         $self->_year_increment($mon);
         # convert to unix time
-        my $time = str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
+        my $time = $self->str2time($5,$4,$3,$2,$mon,$self->{year}-1900,$self->{GMT});
 	my $text = $6;
         $text =~ /^
             \[(.*?)\]        # program   -- 1
@@ -291,7 +364,7 @@ use File::Tail;
 use Getopt::Long;
 use POSIX 'setsid';
 
-my $VERSION = "1.12";
+my $VERSION = "1.13";
 
 # config
 my $rrdstep = 60;
@@ -341,7 +414,7 @@ sub usage
 	print "  --daemon-rrd=DIR   write RRDs to DIR instead of /var/log\n";
 	print "  --daemon-log=FILE  write verbose-log to FILE instead of /var/log/mailgraph.log\n";
 	print "  --ignore-localhost ignore mail to/from localhost (used for virus scanner)\n";
-	print "  --ignore-host=HOST ignore mail to/from HOST (used for virus scanner)\n";
+	print "  --ignore-host=HOST ignore mail to/from HOST regexp (used for virus scanner)\n";
 	print "  --only-mail-rrd    update only the mail rrd\n";
 	print "  --only-virus-rrd   update only the virus rrd\n";
 	print "  --rrd-name=NAME    use NAME.rrd and NAME_virus.rrd for the rrd files\n";
@@ -355,7 +428,7 @@ sub main
 {
 	Getopt::Long::Configure('no_ignore_case');
 	GetOptions(\%opt, 'help|h', 'cat|c', 'logfile|l=s', 'logtype|t=s', 'version|V',
-		'year|y=i', 'host=s', 'verbose|v+', 'daemon|d!',
+		'year|y=i', 'host=s', 'verbose|v', 'daemon|d!',
 		'daemon_pid|daemon-pid=s', 'daemon_rrd|daemon-rrd=s',
 		'daemon_log|daemon-log=s', 'ignore-localhost!', 'ignore-host=s',
 		'only-mail-rrd', 'only-virus-rrd', 'rrd_name|rrd-name=s',
@@ -364,7 +437,7 @@ sub main
 	usage if $opt{help};
 
 	if($opt{version}) {
-		print "mailgraph $VERSION by dws\@ee.ethz.ch\n";
+		print "mailgraph $VERSION by david\@schweikert.ch\n";
 		exit;
 	}
 
@@ -512,7 +585,7 @@ sub process_line($)
 			}
 		}
 		elsif($prog eq 'smtpd') {
-			if($text =~ /^[0-9A-F]+: client=(\S+)/) {
+			if($text =~ /^[0-9A-Z]+: client=(\S+)/) {
 				my $client = $1;
 				return if $opt{'ignore-localhost'} and
 					$client =~ /\[127\.0\.0\.1\]$/;
@@ -520,13 +593,13 @@ sub process_line($)
 					$client =~ /$opt{'ignore-host'}/oi;
 				event($time, 'received');
 			}
-			elsif($opt{'virbl-is-virus'} and $text =~ /^(?:[0-9A-F]+: |NOQUEUE: )?reject: .*: 554.* blocked using irbl.dnsbl.bit.nl/) {
+			elsif($opt{'virbl-is-virus'} and $text =~ /^(?:[0-9A-Z]+: |NOQUEUE: )?reject: .*: 554.* blocked using virbl.dnsbl.bit.nl/) {
 				event($time, 'virus');
 			}
-			elsif($opt{'rbl-is-spam'} and $text    =~ /^(?:[0-9A-F]+: |NOQUEUE: )?reject: .*: 554.* blocked using/) {
+			elsif($opt{'rbl-is-spam'} and $text    =~ /^(?:[0-9A-Z]+: |NOQUEUE: )?reject: .*: 554.* blocked using/) {
 				event($time, 'spam');
 			}
-			elsif($text =~ /^(?:[0-9A-F]+: |NOQUEUE: )?reject: /) {
+			elsif($text =~ /^(?:[0-9A-Z]+: |NOQUEUE: )?reject: /) {
 				event($time, 'rejected');
 			}
 		}
@@ -536,7 +609,7 @@ sub process_line($)
 			}
 		}
 		elsif($prog eq 'cleanup') {
-			if($text =~ /^[0-9A-F]+: (?:reject|discard): /) {
+			if($text =~ /^[0-9A-Z]+: (?:reject|discard): /) {
 				event($time, 'rejected');
 			}
 		}
@@ -595,7 +668,7 @@ sub process_line($)
 		}
 	}
 	elsif($prog eq 'amavis' || $prog eq 'amavisd') {
-		if(   $text =~ /^\([0-9-]+\) (Passed|Blocked) SPAM\b/) {
+		if(   $text =~ /^\([0-9-]+\) (Passed|Blocked) SPAM(?:MY)?\b/) {
 			event($time, 'spam'); # since amavisd-new-2004xxxx
 		}
 		elsif($text =~ /^\([0-9-]+\) (Passed|Not-Delivered)\b.*\bquarantine spam/) {
@@ -636,11 +709,11 @@ sub process_line($)
 			event($time, 'virus');
 		}
 		# Vexira antispam
-		if($text =~ /\bcontains spam\b/) {
+		elsif($text =~ /\bcontains spam\b/) {
 			event($time, 'spam');
 		}
 	}
-	elsif($prog eq 'avgatefwd') {
+	elsif($prog eq 'avgatefwd' or $prog eq 'avmailgate.bin') {
 		# AntiVir MailGate
 		if($text =~ /^Alert!/) {
 			event($time, 'virus');
@@ -659,6 +732,10 @@ sub process_line($)
 		if($text =~ /^(?:spamd: )?identified spam/) {
 			event($time, 'spam');
 		}
+		# ClamAV SpamAssassin-plugin
+		elsif($text =~ /(?:result: )?CLAMAV/) {
+			event($time, 'virus');
+		}
 	}
 	elsif($prog eq 'dspam') {
 		if($text =~ /spam detected from/) {
@@ -672,7 +749,7 @@ sub process_line($)
 	}
 	elsif($prog eq 'drweb-postfix') {
 		# DrWeb
-		if($text =~ /infected$/) {
+		if($text =~ /infected/) {
 			event($time, 'virus');
 		}
 	}
@@ -685,17 +762,18 @@ sub process_line($)
 		}
 	}
 	elsif($prog eq 'MailScanner') {
-		if($text =~ /quarantine/ ) {
+		if($text =~ /(Virus Scanning: Found)/ ) {
 			event($time, 'virus');
 		}
-		# you need to turn on "Detail Spam Report = yes" in
-		# MailScanner.conf (Gabriele Oleotti_
-		elsif($text =~ /SpamAssassin/ ) {
-			event($time, 'spam');
-		}
-		if($text =~ /Bounce to/ ) {
+		elsif($text =~ /Bounce to/ ) {
 			event($time, 'bounced');
 		}
+		elsif($text =~ /^Spam Checks: Found ([0-9]+) spam messages/) {
+			my $cnt = $1;
+			for (my $i=0; $i<$cnt; $i++) {
+				event($time, 'spam');
+			}
+		}
 	}
 	elsif($prog eq 'clamsmtpd') {
 		if($text =~ /status=VIRUS/) {
@@ -707,6 +785,12 @@ sub process_line($)
 			event($time, 'virus');
 		}
 	}
+	# uncommment for clamassassin:
+	#elsif($prog eq 'clamd') {
+	#	if($text =~ /^stream: .* FOUND$/) {
+	#		event($time, 'virus');
+	#	}
+	#}
 	elsif ($prog eq 'smtp-vilter') {
 		if ($text =~ /clamd: found/) {
 			event($time, 'virus');
@@ -727,10 +811,15 @@ sub process_line($)
 		}
 	}
 	elsif($prog eq 'filter-module') {
-		if($text =~ /\bspam_status\=yes/) {
+		if($text =~ /\bspam_status\=(?:yes|spam)/) {
 			event($time, 'spam');
 		}
 	}
+	elsif($prog eq 'sta_scanner') {
+		if($text =~ /^[0-9A-F]+: virus/) {
+			event($time, 'virus');
+		}
+	}
 }
 
 sub event($$)
@@ -796,7 +885,7 @@ B<mailgraph> [I<options>...]
  --daemon-rrd=DIR   write RRDs to DIR instead of /var/log
  --daemon-log=FILE  write verbose-log to FILE instead of /var/log/mailgraph.log
  --ignore-localhost ignore mail to/from localhost (used for virus scanner)
- --ignore-host=HOST ignore mail to/from HOST (used for virus scanner)
+ --ignore-host=HOST ignore mail to/from HOST regexp (used for virus scanner)
  --only-mail-rrd    update only the mail rrd
  --only-virus-rrd   update only the virus rrd
  --rrd-name=NAME    use NAME.rrd and NAME_virus.rrd for the rrd files
@@ -826,7 +915,8 @@ Metalog (see http://metalog.sourceforge.net/)
 
 =head1 COPYRIGHT
 
-Copyright (c) 2004-2005 by ETH Zurich. All rights reserved.
+Copyright (c) 2000-2007 by ETH Zurich
+Copyright (c) 2000-2007 by David Schweikert
 
 =head1 LICENSE
 
@@ -846,12 +936,7 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
 =head1 AUTHOR
 
-S<David Schweikert E<lt>dws@ee.ethz.chE<gt>>
-
-=head1 HISTORY
-
- 2002-03-19 ds Version 0.20
- 2004-10-11 ds Initial ISGTC version (1.10)
+S<David Schweikert E<lt>david@schweikert.chE<gt>>
 
 =cut
 
